import os
from sys import version_info as python_version_info


def get_version() -> str:
    """
    Get the version of the application based on the git tag.

    :return: The version of the application.
    """

    if python_version_info.major == 3 and python_version_info.minor >= 8:
        from importlib.metadata import PackageNotFoundError, version

        try:
            return version("cellarium-cas")
        except PackageNotFoundError:
            return os.environ.get("CAS_VERSION", "0.0.1")
    elif python_version_info.major == 3 and python_version_info.minor == 7:
        from importlib_metadata import PackageNotFoundError, version

        try:
            return version("cellarium-cas")
        except PackageNotFoundError:
            return os.environ.get("CAS_VERSION", "0.0.1")
    else:
        raise Exception("Unsupported Python version. Please use Python 3.7 or later.")



from enum import Enum


class HTTP:
    """
    HTTP status codes constants.

    """

    STATUS_200_OK = 200
    STATUS_201_CREATED = 201
    STATUS_202_ACCEPTED = 202
    STATUS_204_NO_CONTENT = 204
    STATUS_226_IM_USED = 226

    STATUS_301_MOVED_PERMANENTLY = 301
    STATUS_308_PERMANENT_REDIRECT = 308

    STATUS_400_BAD_REQUEST = 400
    STATUS_401_UNAUTHORIZED = 401
    STATUS_403_FORBIDDEN = 403
    STATUS_NOT_FOUND = 404

    STATUS_500_INTERNAL_SERVER_ERROR = 500
    STATUS_503_SERVICE_UNAVAILABLE = 503
    STATUS_504_GATEWAY_TIMEOUT = 504
    STATUS_511_NETWORK_AUTHENTICATION_REQUIRED = 511


class CountMatrixInput(Enum):
    """
    Constants for the count matrix input type.

    """

    X: str = "X"
    RAW_X: str = "raw.X"


class Headers:
    """
    Header constants that are potentially sent to the CAS API.

    """

    # The authorization header.
    authorization = "Authorization"
    # The client session id that is used to track a user's CAS client session.
    client_session_id = "x-client-session-id"
    # The client action id that is used to track a user's logical action that may span multiple requests.
    client_action_id = "x-client-action-id"


class CellMetadataFeatures(Enum):
    """
    Represents the cell features that can be queried for in the CAS API.

    """

    CAS_CELL_INDEX: str = "cas_cell_index"
    CELL_TYPE: str = "cell_type"
    ASSAY: str = "assay"
    DISEASE: str = "disease"
    DONOR_ID: str = "donor_id"
    IS_PRIMARY_DATA: str = "is_primary_data"
    DEVELOPMENT_STAGE: str = "development_stage"
    ORGANISM: str = "organism"
    SELF_REPORTED_ETHNICITY: str = "self_reported_ethnicity"
    SEX: str = "sex"
    SUSPENSION_TYPE: str = "suspension_type"
    TISSUE: str = "tissue"
    TOTAL_MRNA_UMIS: str = "total_mrna_umis"
    CELL_TYPE_ONTOLOGY_TERM_ID: str = "cell_type_ontology_term_id"
    ASSAY_ONTOLOGY_TERM_ID: str = "assay_ontology_term_id"
    DISEASE_ONTOLOGY_TERM_ID: str = "disease_ontology_term_id"
    DEVELOPMENT_STAGE_ONTOLOGY_TERM_ID: str = "development_stage_ontology_term_id"
    ORGANISM_ONTOLOGY_TERM_ID: str = "organism_ontology_term_id"
    SELF_REPORTED_ETHNICITY_ONTOLOGY_TERM_ID: str = "self_reported_ethnicity_ontology_term_id"
    SEX_ONTOLOGY_TERM_ID: str = "sex_ontology_term_id"
    TISSUE_ONTOLOGY_TERM_ID: str = "tissue_ontology_term_id"



import typing as t

import anndata
import numpy as np
import pandas as pd
import scipy.sparse as sp

from cellarium.cas import constants
from cellarium.cas.preprocessing import callbacks


def _get_adata_var_index_or_by_column(adata: anndata.AnnData, var_column_name: str) -> t.List[str]:
    if var_column_name == "index":
        return adata.var.index.tolist()

    return adata.var[var_column_name].values.tolist()


def pre_sanitize(adata: anndata.AnnData, count_matrix_input: constants.CountMatrixInput) -> None:
    """
    Apply preprocessing steps that are necessary before validation and sanitization
    """
    callbacks.pre_sanitize_callback(adata=adata, count_matrix_input=count_matrix_input)


def sanitize(
    adata: anndata.AnnData,
    cas_feature_schema_list: t.List[str],
    count_matrix_input: constants.CountMatrixInput,
    feature_ids_column_name: str,
    feature_names_column_name: t.Optional[str] = None,
) -> anndata.AnnData:
    """
    Cellarium CAS sanitizing script. Returns a new `anndata.AnnData` instance, based on the feature expression
    matrix of the input instance. Extra features get omitted. Missing features get filled with zeros. If the
    data matrix is not of type float32, it will be converted to float32.

    :param adata: Instance to sanitize
    :param cas_feature_schema_list: List of Ensembl feature ids to rely on.
    :param count_matrix_input: Where to obtain a feature expression count matrix from. Choice of: 'X', 'raw.X'
    :param feature_ids_column_name: Column name where to obtain Ensembl feature ids. Default `index`.
    :param feature_names_column_name: Column name where to obtain feature names. If not provided, no feature names
        should be mapped |br|
    `Default`: ``None``
    :return: `anndata.AnnData` instance that is ready to use with CAS
    """
    if feature_ids_column_name != "index" and feature_ids_column_name not in adata.var.columns.values:
        raise ValueError(
            "`feature_ids_column_name` should have a value of either 'index' "
            "or be present as a column in the `adata.var` object."
        )
    if feature_names_column_name not in {"index", None} and feature_names_column_name not in adata.var.columns.values:
        raise ValueError(
            "`feature_ids_name_column_name` should have a value of either 'index' "
            "or be present as a column in the `adata.var` object."
        )

    adata_feature_schema_list = _get_adata_var_index_or_by_column(adata=adata, var_column_name=feature_ids_column_name)
    original_obs_ids = adata.obs.index.values

    cas_feature_schema_set = set(cas_feature_schema_list)
    adata_feature_schema_set = set(adata_feature_schema_list)
    feature_id_intersection = adata_feature_schema_set.intersection(cas_feature_schema_set)

    cas_feature_id_map = {feature_id: index for index, feature_id in enumerate(cas_feature_schema_list)}
    adata_feature_id_map = {feature_id: index for index, feature_id in enumerate(adata_feature_schema_list)}
    feature_id_intersection_cas_indices = list(map(cas_feature_id_map.get, feature_id_intersection))
    feature_id_intersection_adata_indices = list(map(adata_feature_id_map.get, feature_id_intersection))

    n_cells = adata.shape[0]
    n_features = len(cas_feature_schema_list)
    input_matrix = adata.X if count_matrix_input == constants.CountMatrixInput.X else adata.raw.X

    # Translate the columns from one matrix to another, convert to COO format to make this efficient.
    col_trans = np.zeros(n_features, dtype=int)
    for i, k in enumerate(feature_id_intersection_cas_indices):
        col_trans[i] = k
    vals = input_matrix.tocsc()[:, feature_id_intersection_adata_indices]
    vals = vals.tocoo()
    new_col = col_trans[vals.col]
    result_matrix = sp.coo_matrix((vals.data, (vals.row, new_col)), shape=(n_cells, n_features))
    del col_trans, vals, new_col

    # Create `obs` index
    obs = adata.obs.copy()
    obs.index = original_obs_ids

    var_df_data = {}
    if feature_names_column_name is not None:
        adata_feature_name_list = _get_adata_var_index_or_by_column(
            adata=adata, var_column_name=feature_names_column_name
        )
        gene_id_to_gene_symbol_map = {
            gene_id: gene_symbol for gene_symbol, gene_id in zip(adata_feature_name_list, adata_feature_schema_list)
        }

        cas_feature_name_list = [
            gene_id_to_gene_symbol_map[gene_id] if gene_id in gene_id_to_gene_symbol_map else "N/A"
            for gene_id in cas_feature_schema_list
        ]

        var_df_data["feature_name"] = cas_feature_name_list

    var_df = pd.DataFrame(index=cas_feature_schema_list, data=var_df_data)

    return anndata.AnnData(
        result_matrix.tocsr(),
        obs=obs,
        obsm=adata.obsm,
        var=var_df,
    )



import typing as t

from pydantic import BaseModel, Field


class CellTypeSummaryStatisticsResults(BaseModel):
    """
    Represents the data object returned by the CAS API for nearest neighbor annotations.
    """

    class DatasetStatistics(BaseModel):
        dataset_id: str = Field(
            description="The ID of the dataset containing cells", examples=["a7a92fb49-50741b00a-244955d47"]
        )
        count_per_dataset: int = Field(description="The number of cells found in the dataset", examples=[10])
        min_distance: float = Field(
            description="The minimum distance between the query cell and the dataset cells",
            examples=[1589.847900390625],
        )
        max_distance: float = Field(
            description="The maximum distance between the query cell and the dataset cells",
            examples=[1840.047119140625],
        )
        median_distance: float = Field(
            description="The median distance between the query cell and the dataset cells", examples=[1791.372802734375]
        )
        mean_distance: float = Field(
            description="The mean distance between the query cell and the dataset cells", examples=[1791.372802734375]
        )

    class SummaryStatistics(BaseModel):
        cell_type: str = Field(description="The cell type of the cluster of cells", examples=["erythrocyte"])
        cell_count: int = Field(description="The number of cells in the cluster", examples=[94])
        min_distance: float = Field(
            description="The minimum distance between the query cell and the cluster cells",
            examples=[1589.847900390625],
        )
        p25_distance: float = Field(
            description="The 25th percentile distance between the query cell and the cluster cells",
            examples=[1664.875244140625],
        )
        median_distance: float = Field(
            description="The median distance between the query cell and the cluster cells", examples=[1791.372802734375]
        )
        p75_distance: float = Field(
            description="The 75th percentile distance between the query cell and the cluster cells",
            examples=[1801.3585205078125],
        )
        max_distance: float = Field(
            description="The maximum distance between the query cell and the cluster cells",
            examples=[1840.047119140625],
        )
        dataset_ids_with_counts: t.Optional[t.List["CellTypeSummaryStatisticsResults.DatasetStatistics"]] = None

    class NeighborhoodAnnotation(BaseModel):
        """
        Represents the data object returned by the CAS API for a single nearest neighbor annotation.
        """

        query_cell_id: str = Field(description="The ID of the querying cell", examples=["ATTACTTATTTAGTT-12311"])
        matches: t.List["CellTypeSummaryStatisticsResults.SummaryStatistics"]

    data: t.List["CellTypeSummaryStatisticsResults.NeighborhoodAnnotation"] = Field(description="The annotations found")


CellTypeSummaryStatisticsResults.model_rebuild()


class CellTypeOntologyAwareResults(BaseModel):
    """
    Represents the data object returned by the CAS API for a ontology-aware annotations.
    """

    class Match(BaseModel):
        score: float = Field(description="The score of the match", examples=[0.789])
        cell_type_ontology_term_id: str = Field(
            description="The ontology term ID of the cell type for the match", examples=["CL:0000121"]
        )
        cell_type: str = Field(description="The cell type of the match", examples=["erythrocyte"])

    class OntologyAwareAnnotation(BaseModel):
        """
        Represents the data object returned by the CAS API for a single ontology-aware annotation.
        """

        query_cell_id: str = Field(description="The ID of the querying cell", examples=["ATTACTTATTTAGTT-12311"])
        matches: t.List["CellTypeOntologyAwareResults.Match"] = Field(
            description="The matches found for the querying cell"
        )
        total_weight: float = Field(description="The total weight of the matches", examples=[11.23232])
        total_neighbors: int = Field(description="The total number of neighbors matched", examples=[1023])
        total_neighbors_unrecognized: int = Field(
            description="The total number of neighbors that were not recognized", examples=[5]
        )

    data: t.List["CellTypeOntologyAwareResults.OntologyAwareAnnotation"] = Field(description="The annotations found")


CellTypeOntologyAwareResults.model_rebuild()


class MatrixQueryResults(BaseModel):
    """
    Represents the data object returned by the CAS API when performing a cell matrix query
    (e.g. a query of the cell database using a matrix).
    """

    class Match(BaseModel):
        cas_cell_index: float = Field(description="CAS-specific ID of a single cell", examples=[123])
        distance: float = Field(
            description="The distance between this querying cell and the found cell", examples=[0.123]
        )

    class MatrixQueryResult(BaseModel):
        """
        Represents the data object returned by the CAS API for a single cell query.
        """

        query_cell_id: str = Field(description="The ID of the querying cell", examples=["ATTACTTATTTAGTT-12311"])
        neighbors: t.List["MatrixQueryResults.Match"]

    data: t.List["MatrixQueryResults.MatrixQueryResult"] = Field(description="The results of the query")


MatrixQueryResults.model_rebuild()


class CellQueryResults(BaseModel):
    """
    Represents the data object returned by the CAS API for a cell query.
    """

    class CellariumCellMetadata(BaseModel):
        cas_cell_index: int = Field(description="The CAS-specific ID of the cell", examples=[123])
        cell_type: t.Optional[str] = Field(description="The cell type of the cell", examples=["enterocyte"])
        assay: t.Optional[str] = Field(description="The assay used to generate the cell", examples=["10x 3' v2"])
        disease: t.Optional[str] = Field(description="The disease state of the cell", examples=["glioblastoma"])
        donor_id: t.Optional[str] = Field(description="The ID of the donor of the cell", examples=["H20.33.013"])
        is_primary_data: t.Optional[bool] = Field(description="Whether the cell is primary data", examples=[True])
        development_stage: t.Optional[str] = Field(
            description="The development stage of the cell donor", examples=["human adult stage"]
        )
        organism: t.Optional[str] = Field(description="The organism of the cell", examples=["Homo sapiens"])
        self_reported_ethnicity: t.Optional[str] = Field(
            description="The self reported ethnicity of the cell donor", examples=["Japanese"]
        )
        sex: t.Optional[str] = Field(description="The sex of the  cell donor", examples=["male"])
        suspension_type: t.Optional[str] = Field(description="The cell suspension types used", examples=["nucleus"])
        tissue: t.Optional[str] = Field(
            description="The tissue-type that the cell was a part of", examples=["cerebellum"]
        )
        total_mrna_umis: t.Optional[int] = Field(
            description="The count of mRNA UMIs associated with this cell", examples=[24312]
        )

        # Ontology term IDs for the fields
        cell_type_ontology_term_id: t.Optional[str] = Field(
            description="The ID used by the ontology for the type of the cell", examples=["CL:0000121"]
        )
        assay_ontology_term_id: t.Optional[str] = Field(
            description="The ID used by the ontology for the assay used to generate the cell", examples=["EFO:0010550"]
        )
        disease_ontology_term_id: t.Optional[str] = Field(
            description="The ID used by the ontology for the disease state of the cell", examples=["PATO:0000461"]
        )
        development_stage_ontology_term_id: t.Optional[str] = Field(
            description="The ID used by the ontology for the development stage of the cell donor",
            examples=["HsapDv:0000053"],
        )
        organism_ontology_term_id: t.Optional[str] = Field(
            description="The ID used by the ontology for the organism of the cell", examples=["NCBITaxon:9606"]
        )
        self_reported_ethnicity_ontology_term_id: t.Optional[str] = Field(
            description="The ID used by the ontology for the self reported ethnicity of the cell donor",
            examples=["HANCESTRO:0019"],
        )
        sex_ontology_term_id: t.Optional[str] = Field(
            description="The ID used by the ontology for the sex of the cell donor", examples=["PATO:0000384"]
        )
        tissue_ontology_term_id: t.Optional[str] = Field(
            description="The ID used by the ontology for the tissue type that the cell was a part of",
            examples=["UBERON:0002037"],
        )

    data: t.List["CellQueryResults.CellariumCellMetadata"] = Field(description="The metadata of the found cells")


CellQueryResults.model_rebuild()
